
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>

#include <gtk/gtkmain.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkselection.h>

#include "ansiterm.h"

void child_signal_handler(int signum);

int pty_open(char *command, int *childpid, int col, int row);
void pty_setwinsize(int fd, int col, int row);

GdkColor **color_alloc(void);
int send_to_pty(char *buf, int bytes);
void input_handler(gpointer data, int source, GdkInputCondition con);
void check_in_selection(AnsiTerm *term);
static void terminal_size_changed(AnsiTerm *term);
static void parse_command_line_option(int *argc, char **argv[]);
static void set_term_background(AnsiTerm *term);



pid_t childpid;
int pos_x, pos_y;
int shell_fd;
AnsiTerm *term;
GdkPixmap *bg_pixmap = NULL;

/* variable for command line option */
char *font_name = NULL;
char *geometry = NULL;
char *fg_color = NULL;
char *bg_color = NULL;

gint
move_resize_cb(GtkWidget *w, int *x, int *y, int width, int height)
{
	//printf("%d %d %d %d\n", *x, *y, width, height);
	//printf("%d %d\n", w->requisition.width, w->requisition.height);
	return TRUE;
}

int
main(int argc, char *argv[])
{
	char command[256];
	GdkColor **color_table;
	char *s;
	int col, row, saved_lines;
	int terminal_font_ascent, terminal_font_height;
	GtkWidget *win, *hbox;

	terminal_font_ascent = -1;
	terminal_font_height = -1;

	if(!font_name) 
		font_name = "-schumacher-clean-medium-r-normal--16-160-*-irv*,hy16";

	col = 80;
	row = 25;

	saved_lines = 400;


	gtk_init(&argc, &argv);

	parse_command_line_option(&argc, &argv);

	strcpy(command,getenv("SHELL"));
	strcat(command, " -l");
	if(geometry) {
		//printf("geometry: %s\n", geometry);
		s = geometry;
		if(isdigit(*s)) {
			col = *s++ - '0';
			while(isdigit(*s)) {
				col = col * 10 + *s++ - '0';
			}
			if(*s++ != 'x') {
				g_error("invalid -geometry option1");
			}
			row = 0;
			while(isdigit(*s)) {
				row = row * 10 + *s++ - '0';
			}
		}
		if(*s == '+') {
			pos_x = *++s - '0';
			while(isdigit(*s)) {
				pos_x = pos_x + *s++ - '0';
			}
		} else if( *s == '-') {
			pos_x = *++s - '0';
			while(isdigit(*s)) {
				pos_x = pos_x + *s++ - '0';
			}
			pos_x = -pos_x;
		} else if(*s != 0) {
			g_error("invalid -geometry option2");
		}
		if(*s == '+') {
			pos_y = *++s - '0';
			while(isdigit(*s)) {
				pos_y = pos_y + *s++ - '0';
			}
		} else if( *s == '-') {
			pos_y = *++s - '0';
			while(isdigit(*s)) {
				pos_y = pos_y + *s++ - '0';
			}
			pos_y = -pos_y;
		} else if(*s != 0) {
			g_error("invalid -geometry option");
		}
	}


	win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_realize(win);
	gtk_signal_connect(GTK_OBJECT(win), "move_resize",
		GTK_SIGNAL_FUNC(move_resize_cb), NULL);
	gtk_window_set_policy(GTK_WINDOW(win), TRUE, TRUE, TRUE);
	gtk_container_border_width(GTK_CONTAINER(win), 5);

	hbox = gtk_hbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(win), hbox);
	//gtk_container_border_width(GTK_CONTAINER(hbox), 5);
	gtk_widget_show(hbox);


	color_table = color_alloc();

	term = (AnsiTerm *)ansi_term_new(font_name, col, row, saved_lines, terminal_font_ascent, terminal_font_height, NULL, send_to_pty, color_table);

	if(term == NULL) {
		g_warning("fail to create terminal");
		return 1;
	}
	gtk_box_pack_start(GTK_BOX(hbox), (GtkWidget*)term, TRUE, TRUE, 0);
	gtk_widget_show((GtkWidget*)term);
	gtk_widget_grab_focus(GTK_WIDGET(term));

	gtk_signal_connect(GTK_OBJECT(term), "received",
		GTK_SIGNAL_FUNC(check_in_selection), NULL);
	gtk_signal_connect(GTK_OBJECT(term), "button_press_event",
		GTK_SIGNAL_FUNC(button_press_callback), NULL);
	gtk_signal_connect(GTK_OBJECT(term), "button_release_event",
		GTK_SIGNAL_FUNC(button_release_callback), NULL);
	gtk_signal_connect(GTK_OBJECT(term), "motion_notify_event",
		GTK_SIGNAL_FUNC(motion_callback), NULL);
	gtk_signal_connect(GTK_OBJECT(term), "size_changed",
		GTK_SIGNAL_FUNC(terminal_size_changed), NULL);
	gtk_signal_connect(GTK_OBJECT(term), "selection_clear_event",
		GTK_SIGNAL_FUNC(ansi_term_selection_clear), NULL);
	gtk_signal_connect(GTK_OBJECT(term), "realize",
		GTK_SIGNAL_FUNC(set_term_background), NULL);
	gtk_selection_add_handler(GTK_WIDGET(term), GDK_SELECTION_PRIMARY,
		GDK_TARGET_STRING, ansi_term_selection_handler, NULL, NULL);

	keymap_init(term);

	gtk_signal_connect(GTK_OBJECT(term->adjustment), "value_changed",
		GTK_SIGNAL_FUNC(scrollbar_move_callback), term);
	term->scrollbar = gtk_vscrollbar_new(term->adjustment);
	gtk_box_pack_end(GTK_BOX(hbox), term->scrollbar, FALSE, FALSE, 0);
	gtk_widget_show(term->scrollbar);

	signal(SIGCHLD, child_signal_handler);
	shell_fd = pty_open(command, &childpid, col, row);
	if(shell_fd < 0) {
		g_warning("fail to execute shell");
		return 1;
	}
	gdk_input_add(shell_fd, GDK_INPUT_READ, input_handler, NULL);



	gtk_widget_show(win);
	gtk_main();
	return 0;
}

void
input_handler(gpointer data, int source, GdkInputCondition con)
{
	int size;
	char buf[256];
	size = read(source, buf, 255);
	if(size)term->to_term(term, buf, size);
}

void
child_signal_handler(int signum)
{
	if(wait(NULL) == childpid) {
		gtk_main_quit();
		return;
	}
	signal(SIGCHLD, child_signal_handler);
}

GdkColor *
color_new(float red, float green, float blue)
{
	GdkColor *color;
	color = g_new(GdkColor, 1);
	color->red = red * 65535.0;
	color->green = green * 65535.0;
	color->blue = blue * 65535.0;
	return color;
}

GdkColor *black = NULL;
GdkColor *red = NULL;
GdkColor *green = NULL;
GdkColor *yellow = NULL;
GdkColor *blue = NULL;
GdkColor *magenta = NULL;
GdkColor *cyan = NULL;
GdkColor *white = NULL;

GdkColor **
color_alloc(void)
{
	GdkColormap *cmap;
	GdkColor **color_table;
	int i;

	color_table = g_new(GdkColor *, 9);
	cmap = gdk_colormap_get_system();

	if(!black) black = color_new(0, 0, 0);
	if(!red) red = color_new(0.7, 0, 0);
	if(!green) green = color_new(0, 0.7, 0);
	if(!yellow) yellow = color_new(0.6, 0.6, 0);
	if(!blue) blue = color_new(0.1, 0.1, 0.5);
	if(!magenta) magenta = color_new(0.5, 0, 0.5);
	if(!cyan) cyan = color_new(0, 0.6, 0.6);
	if(!white) white = color_new(0.78, 0.78, 0.80);

	color_table[TERM_COLOR_BLACK] = black;
	color_table[TERM_COLOR_RED] = red;
	color_table[TERM_COLOR_GREEN] = green;
	color_table[TERM_COLOR_YELLOW] = yellow;
	color_table[TERM_COLOR_BLUE] = blue;
	color_table[TERM_COLOR_MAGENTA] = magenta;
	color_table[TERM_COLOR_CYAN] = cyan;
	color_table[TERM_COLOR_WHITE] = white;

	color_table[TERM_COLOR_I_SELECT] = color_new(0.9, 0.5, 0.7);

	for(i=0;i<9;i++) gdk_color_alloc(cmap, color_table[i]);

	return color_table;
}

int
send_to_pty(char *buf, int bytes)
{
	write(shell_fd, buf, bytes);
	return TRUE;
}

static void
terminal_size_changed(AnsiTerm *term)
{
	//g_print("%s %d %d %d %d\n", __FUNCTION__, term->col, term->row, term->bot, term->top);
	pty_setwinsize(shell_fd, term->col, term->row);
}

static void
parse_command_line_option(int *argc, char ***argv)
{
	int i, j, k;

	for (i = 1; i< *argc; ) {
		if (strcmp ("-e", (*argv)[i]) == 0) {
			break;
		} else if (strcmp ("-fn", (*argv)[i]) == 0) {
			(*argv)[i] = NULL;
			if ((i + 1) < *argc) {
				font_name = g_strdup((*argv)[i+1]);
				(*argv)[i + 1] = NULL;
			}
			i += 2;
		} else if (strcmp ("-geometry", (*argv)[i]) == 0) {
			(*argv)[i] = NULL;
			if ((i + 1) < *argc) {
				geometry = g_strdup((*argv)[i + 1]);
				(*argv)[i + 1] = NULL;
			}
			i += 2;
		} else if (strcmp ("-bg", (*argv)[i]) == 0) {
			(*argv)[i] = NULL;
			if ((i + 1) < *argc) {
				bg_color = g_strdup((*argv)[i + 1]);
				(*argv)[i + 1] = NULL;
			}
			i += 2;
		} else if (strcmp ("-fg", (*argv)[i]) == 0) {
			(*argv)[i] = NULL;
			if ((i + 1) < *argc) {
				fg_color = g_strdup((*argv)[i + 1]);
				(*argv)[i + 1] = NULL;
			}
			i += 2;
		}
	}

	for (i = 1; i < *argc; i++) {
		for (k = i; k < *argc; k++)
			if ((*argv)[k] != NULL)
				break;

		if (k > i) {
			k -= i;
			for (j = i + k; j < *argc; j++)
				(*argv)[j-k] = (*argv)[j];
			*argc -= k;
		}
	}
}

static void
set_term_background(AnsiTerm *term)
{
	GtkWidget *widget;

	widget = (GtkWidget*) term;

	if(bg_pixmap)
		gdk_window_set_back_pixmap(widget->window, bg_pixmap, FALSE);
	else
		gdk_window_set_background(widget->window, term->color_table[term->normal_bg]);
}

